/*
 * This file is the entry point for the main thread
 * It takes care of spawning the worker and sending it the initial init message
 * It also acts and the messenger between the worker thread and the renderer running on the UI
 * thread
*/
library angular2.src.web_workers.ui.impl;

import "di_bindings.dart" show createInjector;
import "package:angular2/src/web_workers/shared/message_bus.dart"
    show MessageBus, MessageBusSink;
import "package:angular2/src/core/application_common.dart" show createNgZone;
import "package:angular2/di.dart" show Injectable;
import "package:angular2/src/dom/browser_adapter.dart" show BrowserDomAdapter;
import "package:angular2/src/profile/wtf_init.dart" show wtfInit;
import "package:angular2/src/web_workers/ui/setup.dart" show WebWorkerSetup;
import "package:angular2/src/web_workers/ui/render_compiler.dart"
    show MessageBasedRenderCompiler;
import "package:angular2/src/web_workers/ui/renderer.dart"
    show MessageBasedRenderer;
import "package:angular2/src/web_workers/ui/xhr_impl.dart"
    show MessageBasedXHRImpl;

/**
 * Creates a zone, sets up the DI bindings
 * And then creates a new WebWorkerMain object to handle messages from the worker
 */
bootstrapUICommon(MessageBus bus) {
  BrowserDomAdapter.makeCurrent();
  var zone = createNgZone();
  wtfInit();
  zone.run(() {
    var injector = createInjector(zone, bus);
    // necessary to kick off all the message based components
    injector.get(WebWorkerMain);
  });
}

@Injectable()
class WebWorkerMain {
  MessageBasedRenderCompiler renderCompiler;
  MessageBasedRenderer renderer;
  MessageBasedXHRImpl xhr;
  WebWorkerSetup setup;
  WebWorkerMain(this.renderCompiler, this.renderer, this.xhr, this.setup) {}
}
